
var Proteus = require("proteus"),
    FS      = require("fs"),
    Path    = require("path"),
    Glob    = require("glob"),
    nutil   = require("util"),
    sutil   = require("sake/util"),
    Task    = require("sake/task"),
    log     = require("sake/logger"),
    existsSync = FS.existsSync || Path.existsSync,
    SAKEFILE_GLOB_PATTERN = "[Ss]akefile{.js,.coffee,}"
;

module.exports = {
    
    run: function () {
        var opts    = this.options,
            cmd     = opts.taskname,
            args    = opts.arguments,
            task
        ;
        
        // log.debug(opts);
        
        this.loadSakefile(opts.sakefile);
        
        this.processRunOptions();
        
        Task.invoke.apply(Task, [opts.taskname].concat(args));
    },
    
    processRunOptions: function () {
        var opts    = this.options,
            pattern = opts.taskname,
            exit    = false;
        
        pattern = pattern === "default" ? false : new RegExp(pattern);
        
        if (opts.tasks) {
            this.listTasks(pattern);
            exit = true;
        }
        else if (opts.prereqs) {
            this.listPrerequisites(pattern);
            exit = true;
        }
        
        if (exit) {
            process.exit(0);
        }
    },
    
    loadSakefile: function (filepath) {
        var opts = this.options;
        
        if (!opts.nosearch) {
            filepath = filepath || this.sakefileLocation();

            if (!existsSync(filepath)) {
                throw new Error("Can not find a Sakefile.");
            }

            filepath = Path.resolve(filepath);
            opts.sakefile = filepath;
        }
        
        if (opts.nochdir) {
            log.info("sake file: " + filepath);
        }
        else {
            process.chdir(Path.dirname(filepath));
        }

        log.info("sake in " + process.cwd());

        log.debug(require.resolve("sake"));
        this.sake = require("sake");
        if (filepath) {
            this.sake.load(filepath);
        }
    },
    
    sakefileLocation: (function () {
        var sakefile;
        
        function haveSakefile (dir) {
            return (sakefile = this.haveSakefile(dir));
        }
        
        return function () {
            var opts    = this.options,
                here    = process.cwd(),
                start   = here,
                dirs    = [here],
                pathDirs = [],
                found
            ;

            while (here !== "/") {
                here = Path.resolve(here, "..");
                dirs.push(here);
            }

            if (process.env.SAKE_PATH && !opts.nosystem) {
                pathDirs.push.apply(pathDirs, process.env.SAKE_PATH.split(":"));
            }

            log.debug("Search paths: \n" + dirs.concat(pathDirs).map(function (p) {
                return (Array(4).join(" ") + "\"" + p + "\"");
            }).join("\n"));

            found = dirs.some(haveSakefile, this) ||
                    (pathDirs.some(haveSakefile, this) && (opts.nochdir = true));

            return sakefile;
        };
    }()),
    
    haveSakefile: function (here) {
        var pattern = Path.join(here, SAKEFILE_GLOB_PATTERN),
            files = Glob.sync(pattern)
        ;
        return files && files[0];
    },
    
    listTasks: function (pattern) {
        var nameW = 0;
        
        log("Tasks:");
        
        Task.getAll().filter(function (task) {
            return task.description && (!pattern || pattern.test(task.fqn));
        }).map(function (task) {
            var name = task.fqn, len  = name.length;
            nameW = len > nameW ? len : nameW;
            return task;
        }).forEach(function (t) {
            var name = t.fqn, len  = name.length;
            
            log("  " + name + Array(nameW - len + 5).join(" ") + "# " +
                t.description);
        });
    },
    
    listPrerequisites: (function () {
        
        function formatTask (task, pad) {
            return (pad ? pad : "") + task.fqn + " (" + task.type + ")";
        }
        
        return function (pattern) {
            log("Tasks and Prerequisites: ");
            
            Task.getAll().filter(function (task) {
                return !pattern || pattern.test(task.fqn);
            }).forEach(function (task) {
                var preqs       = task.prerequisites,
                    pad         = "  "
                ;

                log(formatTask(task, pad));
                preqs.forEach(function (name) {
                    log(formatTask(Task.lookup(name, task.namespace), pad + pad));
                });
            });
        };
    }()),
    
    get options () {
        return require("sake/options");
    }
};
