
var Proteus     = require("proteus"),
    Path        = require("path"),
    FileTask    = require("sake/file-task"),
    getOwnPropSpec = Object.getOwnPropertyDescriptor
;

/**
 * A FileCreationTask is a file task that when used as a dependency will
 * be needed if and only if the file has not been created.  Once created,
 * it is not re-triggered if any of its dependencies are newer, nor does
 * it trigger any rebuilds of tasks that depend on it whenever it is
 * updated.
 */
module.exports = FileTask.derive((function () {
    var proto = {},
        isNeededSpec = getOwnPropSpec(FileTask.prototype, "isNeeded")
    ;
    
    proto.type = "file-create-task";
    
    Object.defineProperties(proto, {
        /**
         * @property isNeeded
         * @type {boolean}
         * @override Task#isNeeded
         */
        isNeeded: Proteus.merge(isNeededSpec, {
            get: function () {
                return !this.exists;
            }
        }),
        /**
         * A FileCreateTask timestamp is always the distant past
         * @property timestamp
         * @type {integer}
         * @override Task#timestamp
         */
        timestamp: {
            value: 0,
            enumerable: true
        }
    });
    
    return proto;
}()));